@@ -30,8 +30,10 @@
 #include <gjs/gjs.h>
 
 static char **include_path = NULL;
+static char *command = NULL;
 
 static GOptionEntry entries[] = {
+    { "command", 'c', 0, G_OPTION_ARG_STRING, &command, "Program passed in as a string", "COMMAND" },
     { "include-path", 'I', 0, G_OPTION_ARG_STRING_ARRAY, &include_path, "Add the directory DIR to the list of directories to search for js files.", "DIR" },
     { NULL }
 };
@@ -39,6 +41,7 @@ static GOptionEntry entries[] = {
 int
 main(int argc, char **argv)
 {
+    char *command_line;
     GOptionContext *context;
     GError *error = NULL;
     GjsContext *js_context;
@@ -47,6 +50,8 @@ main(int argc, char **argv)
     gsize len;
     int code;
 
+    g_thread_init(NULL);
+
     context = g_option_context_new(NULL);
 
     /* pass unknown through to the JS script */
@@ -59,6 +64,10 @@ main(int argc, char **argv)
     setlocale(LC_ALL, "");
     g_type_init();
 
+    command_line = g_strjoinv(" ", argv);
+    gjs_debug(GJS_DEBUG_LOG, "Command line: %s", command_line);
+    g_free(command_line);
+
     gjs_debug(GJS_DEBUG_CONTEXT,
               "Creating new context to eval console script");
     js_context = gjs_context_new_with_search_path(include_path);
@@ -69,9 +78,13 @@ main(int argc, char **argv)
                                          &error)) {
         g_printerr("Failed to defined ARGV: %s", error->message);
         exit(1);
-    } 
+    }
 
-    if (argc <= 1) {
+    if (command != NULL) {
+        script = command;
+        len = strlen(script);
+        filename = "<command line>";
+    } else if (argc <= 1) {
         script = g_strdup("const Console = imports.console; Console.interact();");
         len = strlen(script);
         filename = "<stdin>";
@@ -92,7 +105,7 @@ main(int argc, char **argv)
         g_printerr("%s\n", error->message);
         exit(1);
     }
-    
+
     gjs_memory_report("before destroying context", FALSE);
     g_object_unref(js_context);
     gjs_memory_report("after destroying context", TRUE);
