@@ -43,7 +43,7 @@ gjs_main_loop_quit(JSContext *context,
 {
     char *cancel_id;
     GMainLoop *main_loop;
-    
+
     if (!gjs_parse_args(context, "quit", "s", argc, argv,
                         "cancelId", &cancel_id))
       return JS_FALSE;
@@ -123,6 +123,7 @@ closure_source_func(void *data)
         /* closure is invalid now */
         return FALSE;
     }
+    JS_BeginRequest(context);
 
     retval = JSVAL_VOID;
     JS_AddRoot(context, &retval);
@@ -135,12 +136,13 @@ closure_source_func(void *data)
      * JavaScript always makes some sense of any value in
      * an "if (value) {}" context.
      */
-    if (!JS_ValueToBoolean(gjs_closure_get_context(closure),
+    if (!JS_ValueToBoolean(context,
                            retval, &bool_val))
         bool_val = FALSE;
 
     JS_RemoveRoot(context, &retval);
 
+    JS_EndRequest(context);
     return bool_val;
 }
 
@@ -269,7 +271,7 @@ gjs_idle_add(JSContext *context,
     GClosure *closure;
     guint id;
     int priority = G_PRIORITY_DEFAULT_IDLE;
-    
+
     /* Best I can tell, there is no way to know if argv[0] is really
      * callable other than to just try it. Checking whether it's a
      * function will not detect native objects that provide
